package tauzaman.calendricsystem.granularitylattice.minpriorityqueue;

/**
  * <code>MinPriorityQueue</code> is a generic data structure to implement a
  *   minimum priority queue.  The implementation is done with a binary heap
  *   backed by an array.
  *
  * <p>Title: Minimum Priority Queue </p>
  * <p>Description: Generic data structure to implement a minimum priority
  *                 queue </p>
  * <p>Copyright: Copyright (c) 2003</p>
  * @author Jessica Miller
  * @version 1.0
  * @status implementation complete
  **/

import java.util.*;

public class MinPriorityQueue<P extends Comparable<P>, E> {

  /**
    * This is the starting size of the minimum priority queue's array.
    */
  private static final int DEFAULT_SIZE = 10;

  /**
    * This is the index of the root.  It is 1 because the arithmetic for the
    *   queue instructions is much cleaner if the root is at index 1.  This
    *   does have the downside of one wasted element (that element at index 0 in
    *   the array).
    */
  private static final int ROOT_INDEX   = 1;

  /**
    * The array that contains the queue's elements.
    */
  private ArrayList<MPQElement<P, E>> qItems;

  /**
    * This hashtable maps each element's value to the index in the qItems array
    *   in which it resides.  This is needed to maintain quick access to each
    *   element when you want to decrease it's priority.  Otherwise, users would
    *   have to keep track of different elements' priorities in order to make a
    *   change in the element's priority.
    */
  private Hashtable<E, Integer> valueToIndexTable;

  /**
    * The current size of the queue.
    */
  private int qSize;

  /**
    * Constructs a minimum priority queue by initializing its array to the
    *   default size and intializing the value -> index hashtable.
    */
  public MinPriorityQueue()
  {
    qItems            = new ArrayList<MPQElement<P, E>>(DEFAULT_SIZE);
    valueToIndexTable = new Hashtable<E, Integer>();
    qSize  = 0;
    
    // since we don't use first element add place holder
    P dummyPriority = null;
    E dummyElement = null;
    qItems.add(new MPQElement<P, E>(dummyPriority, dummyElement));
  }

  /**
    * Inserts a new element into the priority queue.  First it inserts the
    *   element at the end of the queue (after possibly adjusting the size of
    *   the queue such that the new element will fit).  Second, it bubbles the
    *   element up the heap until it is smaller than its parent.
    *
    * @param priority  The priority of the new element.
    * @param value     The value of the new element.
    */
  public void enqueue(P priority, E value)
  {
    qItems.ensureCapacity(qSize + 1);
    MPQElement<P, E> newItem = new MPQElement<P, E>(priority, value);

    // insert item
    int newItemIndex;

    for (newItemIndex = ++qSize;
         newItemIndex > ROOT_INDEX  &&  priority.compareTo(qItems.get(newItemIndex/2).getPriority()) < 0;
         newItemIndex /= 2)
    {
      setElement(newItemIndex, qItems.get(newItemIndex / 2));
      valueToIndexTable.put(qItems.get(newItemIndex).getValue(),
                            new Integer(newItemIndex));
    }

    setElement(newItemIndex, newItem);

    valueToIndexTable.put(value, new Integer(newItemIndex));
  }

  /**
    * Deletes and returns the element at the root (the element with the minimum
    *   priority) of the heap.  First, save the element at the root to return.
    *   Then delete the last element in the heap and give its value to be the
    *   new root.  Lastly, bubble this element down until its children are both
    *   less than this element or until this element is at a leaf.
    *
    * @return The element with the minimum priority (or null if the queue is
    *           empty.
    */
  public MPQElement<P, E> dequeue()
  {
    if (qSize == 0)
      return null;

    // save copy of element at the root node
    MPQElement<P, E> result = qItems.get(ROOT_INDEX);

    // move last node, node n, in heap to root and delete node n
    setElement(ROOT_INDEX, qItems.get(qSize));
    setElement(qSize, null);

    // decrement size of queue and if the queue is now empty, return the result
    if (--qSize == 0)
      return result;

    // otherwise, do element shifting, if needed
    int index             = ROOT_INDEX;
    MPQElement<P, E> leftChild  = (ROOT_INDEX * 2 <= qSize)
                            ? qItems.get(ROOT_INDEX * 2) : null;
    MPQElement<P, E> rightChild = (ROOT_INDEX * 2 + 1 <= qSize)
                            ? qItems.get(ROOT_INDEX * 2 + 1) : null;
    P priority   = qItems.get(ROOT_INDEX).getPriority();
    MPQElement<P, E> element    = qItems.get(ROOT_INDEX);

    // if new root <= its children, heap exists
    // otherwise, percolate this key down by exchanging it with its smallest
    //   child; repeat until key no longer needs to be moved or is in a leaf
    while (((leftChild != null) && (priority.compareTo(leftChild.getPriority()) > 0))
           || ((rightChild != null) && (priority.compareTo(rightChild.getPriority()) > 0)))
    {
      int childIndex;

      if (leftChild == null)
        childIndex = index * 2 + 1;
      else if (rightChild == null)
        childIndex = index * 2;
      else
        childIndex = (leftChild.getPriority().compareTo(rightChild.getPriority()) <= 0)
                       ? index * 2
                       : index * 2 + 1;

      setElement(index, qItems.get(childIndex));
      valueToIndexTable.put(qItems.get(index).getValue(),
                            new Integer(index));

      index      = childIndex;
      leftChild  = (index * 2 < qSize)
                   ? qItems.get(index * 2) : null;
      rightChild = (index * 2 + 1 < qSize)
                   ? qItems.get(index * 2 + 1) : null;
    }

    setElement(index, element);
    valueToIndexTable.put(element.getValue(), new Integer(index));

    return result;
  }

  /**
    * Given a particular element already in the queue, decrease its priority to
    *   the given priority.
    *
    * @param value        The element whose priority we want to decrease.
    * @param newPriority  The new priority for the element.
    * @return             <code>true</code> if the priority was changed,
    *                       <code>false</code> if the priority given was greater
    *                       than the original priority or if the element is not
    *                       in the queue
    */
  public boolean decreasePriority(E value, P newPriority)
  {
    Object indexOfVal = valueToIndexTable.get(value);
    if (indexOfVal == null)
      return false;

    int valIndex = ((Integer)indexOfVal).intValue();
    MPQElement<P, E> valElement = qItems.get(valIndex);

    if (newPriority.compareTo(valElement.getPriority()) >= 0)
      return false;

    for (;
         valIndex > ROOT_INDEX  &&  newPriority.compareTo(qItems.get(valIndex/2).getPriority()) < 0;
         valIndex /= 2)
    {
        setElement(valIndex, qItems.get(valIndex / 2));
      valueToIndexTable.put(qItems.get(valIndex/2).getValue(),
                            new Integer(valIndex));
    }

    valElement.setPriority(newPriority);
    setElement(valIndex, valElement);
    valueToIndexTable.put(value, new Integer(valIndex));

    return true;
  }

  /**
    * Returns the number of elements currently in the queue.
    *
    * @return  The number of elements currently in the queue.
    */
  public int getSize()
  {
    return qSize;
  }

  /**
    * Returns whether or not the queue is empty.
    *
    * @return  <code>true</code> if the queue is empty, <code>false</code> if
    *            the queue has elements
    */
  public boolean isEmpty()
  {
    return qSize == 0;
  }

  /**
    * Returns a string representation of the minimum priority queue.
    *
    * @return  a string representation of the minimum priority queue
    */
  public String toString()
  {
    int index      = 1;
    int level      = 1;
    int elsAtLevel = 1;
    String res = "";

    while (index <= qSize)
    {
      res += "LEVEL " + level + "\n   ";
      for (int ii = 0; ii < elsAtLevel  &&  index <= qSize; ii++)
      {
        res += qItems.get(index).toString() + "\n   ";
        index++;
      }
      res += "\n";
      elsAtLevel = level * 2;
      level++;
    }

    return res;
  }
  
  /**
   * Sets a specific element of the queue array.  This routine
   * is necessary due to the semantics of the ArrayList class. Since
   * the set() method will cause an error if attempting to set the
   * last element in the array (i.e. index == size).  For this case
   * add() should be used.  This method ensures the correct one of
   * these is invoked.
   * 
   * @param index   The index where to store the element
   * @param element The element to store at specified index
   */
  private void setElement(int index, MPQElement<P, E> element) {
      if (index == qItems.size()){
          // append to end of array
          qItems.add(element);
      } else {
          // replace item at index
          qItems.set(index, element);
      }
  }

  // Simple tests for the minimum priority queue
  public static void main(String[] args)
  {
    MinPriorityQueue<Integer, String> mpq = new MinPriorityQueue<Integer, String>();

    mpq.enqueue(new Integer(225), "Jessica");
    mpq.enqueue(new Integer(2147483647), "Martin");
    mpq.enqueue(new Integer(225), "Meghan");
    mpq.enqueue(new Integer(2147483647), "Laura");

    System.out.println("Before dequeue: " + mpq);

    System.out.println("After dequeued " + mpq.dequeue().getValue());

    System.out.println(mpq);
/*
    mpq.enqueue(new Integer(13), "Jessica");
    mpq.enqueue(new Integer(10), "Martin");
    mpq.enqueue(new Integer(8), "Meghan");
    mpq.enqueue(new Integer(14), "Laura");
    mpq.enqueue(new Integer(15), "John");
    mpq.enqueue(new Integer(9), "Ursula");
    mpq.enqueue(new Integer(4), "Mary");
    mpq.enqueue(new Integer(12), "Brandon");
    mpq.enqueue(new Integer(8), "Chris");
    mpq.enqueue(new Integer(1), "Shannon");

    System.out.println(mpq);

    System.out.print("MinPriorityQueue: [");

    for (int ii = 1;  ii < mpq.getSize(); ii++)
      System.out.print(mpq.qItems.get(ii).getPriority() + ", ");

    System.out.println(mpq.qItems.get(mpq.getSize()).getPriority() + "]");

    mpq.decreasePriority("John", new Integer(3));

    System.out.print("MinPriorityQueue (after decreasing 15's priority to 3): [");

    for (int ii = 1;  ii < mpq.getSize(); ii++)
      System.out.print(mpq.qItems.get(ii).getPriority() + ", ");

    System.out.println(mpq.qItems.get(mpq.getSize()).getPriority() + "]");

    while (mpq.getSize() != 0)
    {
      System.out.println("Element dequeued" + mpq.dequeue());

      System.out.print("MinPriorityQueue (after dequeueing first element): [");

      for (int ii = 1; ii < mpq.getSize(); ii++)
        System.out.print( mpq.qItems.get(ii).getPriority() + ", ");

      if (mpq.getSize() != 0)
        System.out.print( mpq.qItems.get(mpq.getSize()).getPriority());

      System.out.println("]");
    }
    */
  }
}