package tauzaman.extendedboolean;

import java.io.*;

/**
 * <code>ExtendedBoolean</code> class implements the Boolean algebra for a 
 * three-valued logic, with values True, False, and Maybe. In this three-valued 
 * logic, a condition that is known to be True is denoted as
 * True, and a condition that is known to be False is denoted as False. If the
 * condition cannot be determined, either because it lacks information or because of
 * approximations, it is denoted as Maybe.
 * 
 * @author Curtis Dyreson, Alex Henniges, Richard Snodgrass
 * @version 2.0, Sep/2/2008
 * @status IMPLEMENTED
 */
public class ExtendedBoolean implements Serializable {

	private static final long serialVersionUID = 1L;

	// The three types of ExtendedBoolean logic: True, False, and Maybe.
	/**
	 * Enumeration type EB indicates if the <code>ExtendedBoolean</code> is 
	 * True, False, or Maybe
	 */
	public enum LogicValue {
		True, False, Maybe
	};
	
	private LogicValue status;
	
	private static ExtendedBoolean TrueEB = new ExtendedBoolean(LogicValue.True);
	private static ExtendedBoolean FalseEB = new ExtendedBoolean(LogicValue.False);
	private static ExtendedBoolean MaybeEB = new ExtendedBoolean(LogicValue.Maybe);

	private ExtendedBoolean(LogicValue eb) {
		this.status = eb;
	}

	/**
	 * Returns an ExtendedBoolean with the given <code>LogicValue</code>.
	 * 
	 * @param eb
	 * <code>LogicValue</code> enumeration type for the returning ExtendedBoolean to have.
	 * @return
	 * ExtendedBoolean with the given <code>LogicValue</code>.
	 */
	public static ExtendedBoolean valueOf(LogicValue eb) {
		switch(eb) {
		case True: return TrueEB;
		case False: return FalseEB;
		case Maybe: return MaybeEB;
		default: return MaybeEB;
		}
	}

	/**
	 * Logical AND this one with <code>other</code> ExtendedBoolean. The result is False if 
	 * either <code>A</code> or <code>B</code> is False,
     * the result is True if both <code>A</code> and <code>B</code> 
     * are True, else the result is Maybe.
	 * 
	 * @param other -
	 * The other ExtendedBoolean.
	 * @return this AND other.
	 */
	public ExtendedBoolean and(ExtendedBoolean other) {
		switch (status) {
		case True:
			if (other.isSatisfied())
				return TrueEB;
			if (other.isUnsatisfied())
				return FalseEB;
			return MaybeEB;
		case False:
			return FalseEB;
		case Maybe:
		default:
			if (other.isUnsatisfied())
				return FalseEB;
			return this;
		}
	}

	/**
	 * Logical OR this one with <code>other</code> ExtendedBoolean. The result is True if 
	 * either <code>A</code> or <code>B</code> is True,
     * the result is False if both <code>A</code> and <code>B</code>
     * are False, else the result is Maybe.
	 * 
	 * @param other -
	 * the other one
	 * @return this OR other
	 */
	public ExtendedBoolean or(ExtendedBoolean other) {
		switch (status) {
		case True:
			return TrueEB;
		case False:
			if (other.isSatisfied())
				return TrueEB;
			if (other.isUnsatisfied())
				return FalseEB;
			return MaybeEB;
		case Maybe:
		default:
			if (other.isSatisfied())
				return TrueEB;
			return this;
		}
	}

	/**
	 * Logical negation of this ExtendedBoolean. It is 
	 * True if <code>A</code> is False, it is False 
	 * if <code>A</code> is True, else it is Maybe.
	 * 
	 * @return NOT this
	 */
	public ExtendedBoolean not() {
		switch (status) {
		case True:
			return FalseEB;
		case False:
			return TrueEB;
		case Maybe:
		default:
			return this;
		}
	}

	/**
	 * Returns a boolean indicating if this ExtendedBoolean
	 * has a <code>LogicValue</code> of True.
	 * 
	 * @return true if this ExtendedBoolean
	 * has a <code>LogicValue</code> of True.
	 */
	public boolean isSatisfied() {
		return status == LogicValue.True;
	}

	/**
	 * Returns a boolean indicating if this ExtendedBoolean
	 * has a <code>LogicValue</code> of False.
	 * 
	 * @return true if this ExtendedBoolean
	 * has a <code>LogicValue</code> of False.
	 */
	public boolean isUnsatisfied() {
		return status == LogicValue.False;
	}

	/**
	 * Returns a boolean indicating if this ExtendedBoolean
	 * has a <code>LogicValue</code> of Maybe.
	 * 
	 * @return true if this ExtendedBoolean
	 * has a <code>LogicValue</code> of Maybe.
	 */
	public boolean isMaybe() {
		return status == LogicValue.Maybe;
	}
	
	/**
	 * A standard string representation of this ExtendedBoolean's
	 * <code>LogicValue</code>.
	 */
	public String toString() {
		switch(status) {
		case True:
			return "[ExtendedBoolean: True]";
		case False:
			return "[ExtendedBoolean: False]";
		case Maybe:
		default:
			return "[ExtendedBoolean: Maybe]";
		}
	}
}
