package tauzaman.property;

import java.util.*;

import tauzaman.property.Property;



/* Structure: Each property has a corresponding stack. Hashtable maps a given
   property name to its stack. Hence having this hashtable and different stacks
   for different properties not only increase efficiency but also makes the code
   easier.
*/

/**
* <code>PropertyStackService</code> class implements data structures to store properties
* and provides service to <code>PropertyManager</code>.
*
* @author  Curtis Dyreson and Bedirhan Urgun
* @version 0.1, 11/15/02
* @see     tauzaman.property.PropertyManager
* @status design complete, implementation complete
*/

public class PropertyStackService{

  /** Property Stack Hash table, provides efficient indexing of stacks **/
  private Hashtable<String, Stack<Property>> propertyStackMap = null;
  
  /** List of all the property names **/
  private static String [] propertyNames = {"Locale", "InstantInputFormat", "InstantOutputFormat", "IntervalInputFormat",
      "IntervalOutputFormat", "NowRelativeInstantInputFormat", "NowRelativeInstantOutputFormat", "PeriodInputFormat",
      "PeriodOutputFormat", "IndeterminateInstantInputFormat", "IndeterminateInstantOutputFormat", "OverrideInputOrder"};
  
  private static final Integer PROPERTY_COUNT = 12;
    
  /** 
  * Constructs a <code>PropertyStackService</code> object, which provides 
  * stack operations for <code>Property</code>.
  */
  public PropertyStackService(){

      /* form hashtable */
      propertyStackMap = new Hashtable<String, Stack<Property>>();
      initStackMap();
  }

  /** 
  * Initialize the hash table. This might also be done by PropertyManager, 
  * by providing a string[] of property names, however, there are this many
  * properties after all.
  * Each key is the name of the property and correspinding value 
  * is the corresponding index in the property StackList
  */
  public void initStackMap(){

      for (int i = 0; i < PROPERTY_COUNT; i++)
      {
         propertyStackMap.put(propertyNames[i],  new Stack<Property>());
      }
  }
 
  /** 
  * Returns the index of <code>Property</code> given its key, that is to say its name. 
  *
  * @param propertyName name of a <code>Property</code>
  *
  * @return stack, which corresponds to first parameter, propertyName,
  * in stack list
  */
  /*
  private Stack getStack(String propertyName){

      /* validness of propertyName should have already been performed *//*
      int index = ((Integer)propertyStackMap.get(propertyName)).intValue();

      return propertyStackList[index];
  }*/

  /** 
  * Pushes a <code>Property</code> to its stack.
  * @param property a <code>Property</code>
  */
  public void push(Property property){

      Stack<Property> stack = propertyStackMap.get(property.getName());
      /* push this Property into the stack */
      if (stack != null) {
          stack.push(property);
      }
  }

  /** 
  * Pops the corresponding active <code>Property</code> from its stack. Does not
  * deactivates default <code>Property</code>.
  *
  * @param propertyName name of a <code>Property</code>
  */ 
  public void pop(String propertyName){

      Stack<Property> stack = propertyStackMap.get(propertyName);
      /* pop the Property, be careful of default Property */
      if (stack != null)
      {
           if (stack.size() > 1)
               stack.pop();
      }
  }

  /** 
  * Pops all active <code>Property</code>ies from their stacks, except 
  * default ones, which are the last elements in stacks.Does not
  * deactivates default <code>Property</code>.
  */
  public void pop(){
      for (int i = 0; i < PROPERTY_COUNT; i++)
      {
          pop(propertyNames[i]);
      }
  }

  /** 
  * Pops all elements from given property stack until the last elements, which
  * are the default properties.
  *
  * @param propertyName name of a <code>Property</code>
  */
  public void popToLast(String propertyName){
      Stack<Property> stack = propertyStackMap.get(propertyName);
      if (stack != null)
      {
          while(stack.size() > 1)
              stack.pop();
      }
  }

  /** 
  * Pop all elements from their stack until the last elements, which
  * are the default properties. 
  */
  public void popToLast(){
      Stack<Property> stack;
      for (int i = 0; i < PROPERTY_COUNT; i++){
          stack = propertyStackMap.get(propertyNames[i]);
          while(stack.size() > 1)
              stack.pop();
      }
  }

  /** 
  * Gets and returns the top element (<code>Property</code>) of the stack (without removing it).
  * If there is no <code>Property</code>, it returns null.
  *
  * @param propertyName name of a <code>Property</code>
  *
  * @return a <code>Property</code>
  */
  public Property peek(String propertyName){
      Stack<Property> stack = propertyStackMap.get(propertyName);
      if (stack != null) {
            return stack.peek();
        } else {
            return null;
        }
  }

  /**
  * Returns all elements of a given stack
  * 
  * @param propertyName name of a <code>Property</code>
  * @return an Enumeration containing all Elements of a stack
  */
  public Enumeration<Property> peekAll(String propertyName){
      Stack<Property> stack = propertyStackMap.get(propertyName);
        if (stack != null) {
            return stack.elements();
        } else {
            return null;
        }
  }
}
