package tauzaman.temporaldatatypes.semantics;

import tauzaman.calendricsystem.CalendricSystemServiceException;
import tauzaman.calendricsystem.Granularity;
import tauzaman.calendricsystem.granularitylattice.GranularityLattice;
import tauzaman.temporaldatatypes.Instant;
import tauzaman.temporaldatatypes.Interval;
import tauzaman.temporaldatatypes.Period;

/**
 * The <code>CoarserIndeterminateSemantics</code> class performs arithmetic
 * and comparison operations on indeterminate time values. Given two time 
 * values on an operation, the <code>CoarserIndeterminateSemantics</code>
 * converts the time values to the coarser of the two granularities. If the
 * two time values have granularities that cannot be compared directly, then
 * the finest of the coarser granularities common between the two is used. If
 * none exists, an error is thrown. Conversions are performed by a
 * <code>GranularityLattice</code>. A <code>GranularityLattice</code> can be given
 * in the constructor or set before performing operations.
 * 
 * @author  Alex Henniges and Richard Snodgrass
 * @version 2.0, Sep/2/2008
 * @see     tauzaman.temporaldatatypes.Instant
 * @see     tauzaman.calendricsystem.granularitylattice.GranularityLattice
 * @see     tauzaman.temporaldatatypes.semantics.IndeterminateSemantics
 */
public class CoarserIndeterminateSemantics extends IndeterminateSemantics{
	
	/**
	 * Creates a default <code>CoarserIndeterminateSemantics</code> with
	 * <code>GranularityLattice</code> equal to NULL.
	 */
	public CoarserIndeterminateSemantics() {
		lattice = null;
	}
	
	/**
	 * Creates a <code>CoarserIndeterminateSemantics</code> with
	 * <code>GranularityLattice</code> given.
	 * 
	 * @param lattice <code>GranularityLattice</code> for this
	 *  			<code>CoarserIndeterminateSemantics</code>.
	 */
	public CoarserIndeterminateSemantics(GranularityLattice lattice) {
		this.lattice = lattice;
	}

	public void scalePair(Instant left, Instant right) 
	 		throws CalendricSystemServiceException {
		Granularity toGran = lattice.leastUpperBound(left.getGranularity(), right.getGranularity());
		this.leftInstant = left.scale(toGran);
		this.rightInstant = right.scale(toGran);
    }

	public void scalePair(Period left, Period right) 
			throws CalendricSystemServiceException {
		Granularity toGran = lattice.leastUpperBound(left.getGranularity(), right.getGranularity());
		this.leftPeriod = left.scale(toGran);
		this.rightPeriod = right.scale(toGran);
	}
	
	public void scalePair(Period left, Interval right)
			throws CalendricSystemServiceException {
		Granularity toGran = lattice.leastUpperBound(left.getGranularity(), right.getGranularity());
		this.leftPeriod = left.scale(toGran);
		this.rightInterval = right.scale(toGran);
	}
	
	public void scalePair(Interval left, Interval right) 
			throws CalendricSystemServiceException {
		Granularity toGran = lattice.leastUpperBound(left.getGranularity(), right.getGranularity());
		this.leftInterval = left.scale(toGran);
		this.rightInterval = right.scale(toGran);
	}
	
	public void scalePair(Instant left, Interval right) 
			throws CalendricSystemServiceException {
		Granularity toGran = lattice.leastUpperBound(left.getGranularity(), right.getGranularity());
		this.leftInstant = left.scale(toGran);
		this.rightInterval = right.scale(toGran);
	}
	
}
