package tauzaman.temporaldatatypes.semantics;

import tauzaman.calendricsystem.*;
import tauzaman.calendricsystem.granularitylattice.GranularityLattice;
import tauzaman.temporaldatatypes.Instant;
import tauzaman.temporaldatatypes.Interval;
import tauzaman.temporaldatatypes.Period;
import tauzaman.temporaldatatypes.TemporalDataTypeFormationException;
import tauzaman.timestamp.Granule;

public class FinerDeterminateSemantics implements DeterminateSemantics {
	
	private GranularityLattice lattice;
	
	public FinerDeterminateSemantics() {
		lattice = null;
	}
	public FinerDeterminateSemantics(GranularityLattice lattice) {
		this.lattice = lattice;
	}
	
	public void setGranularityLattice(GranularityLattice lattice){
		this.lattice = lattice;
	}
	/** 
	 * Does one Instant precede another, that is, is alpha < beta?
	 * @param alpha - Instant to compare
	 * @param beta - Instant to compare
	 * @return alpha < beta
	 * @throws TemporalDataTypeFormationException 
	 **/
	public boolean precedes(Instant alpha, Instant beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException
	{
		// Determine granularity to convert alpha and/or beta to
		Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());

		Instant gamma = alpha.cast(toGran);
		Instant delta = beta.cast(toGran);
		Granule i = gamma.getGranule();
		Granule j = delta.getGranule();
		return i.compareTo(j) < 0;
	}

	/** 
	 * Does an Instant precede the start of a Period, 
	 * that is, is alpha < beta.start?
	 * @param alpha - Instant to compare
	 * @param beta - Period to compare
	 * @return alpha < beta.start
	 * @throws TemporalDataTypeFormationException 
	 **/
	public boolean precedes(Instant alpha, Period beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
		// Determine granularity to convert alpha and/or beta to
		Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());

		Instant gamma = alpha.cast(toGran);
		Instant delta = beta.getStarting();
		delta = delta.cast(toGran);
		Granule i = gamma.getGranule();
		Granule j = delta.getGranule();
		return i.compareTo(j) < 0;
	}
	
    /** 
	 * Does the end of a Period precede the start of an Instant, 
	 * that is, is alpha.end < beta?
	 * @param alpha - Period to compare
	 * @param beta - Instant to compare
	 * @return alpha.end < beta
     * @throws TemporalDataTypeFormationException 
	 **/
	public boolean precedes(Period alpha, Instant beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException{
		// Determine granularity to convert alpha and/or beta to
		Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());

		Instant gamma = alpha.getEnding();
		gamma = gamma.cast(toGran);
		Instant delta = beta.cast(toGran);
		Granule i = gamma.getGranule();
		Granule j = delta.getGranule();
		return i.compareTo(j) < 0;
	}
	/** 
	 * Does the end of a Period precede the start of another,
	 * that is, is alpha.end < beta.start?
	 * @param alpha - Period to compare
	 * @param beta - Period to compare
	 * @return alpha.end < beta.start
	 * @throws TemporalDataTypeFormationException 
	 **/
	public boolean precedes(Period alpha, Period beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException{
		// Determine granularity to convert alpha and/or beta to
		Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());

		Instant gamma = alpha.getEnding();
		gamma = gamma.cast(toGran);
		Instant delta = beta.getStarting();
		delta = delta.cast(toGran);
		Granule i = gamma.getGranule();
		Granule j = delta.getGranule();
		return i.compareTo(j) < 0;
	}


	  /** 
	   * Does this Period begin at the same time as another, but end before it,
	   * that is, is alpha.start == beta.start but alpha.end < beta.end?  
	   * @param alpha - Period to compare
	   * @param beta - Period to compare
	   * @return alpha.start == beta.start AND alpha.end < beta.end
	   * @throws TemporalDataTypeFormationException 
	   **/
	public boolean starts(Period alpha, Period beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException{
	    // Determine granularity to convert alpha and/or beta to
		Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());
		Period gamma = alpha.cast(toGran);
		Period delta = beta.cast(toGran);
	    Instant gammaStart = gamma.getStarting();
	    Instant gammaEnd = gamma.getEnding();
	    Instant deltaStart = delta.getStarting();
	    Instant deltaEnd = delta.getEnding();

	    // Get both granules
	    Granule i = gammaStart.getGranule();
	    Granule j = deltaStart.getGranule();
	    Granule k = gammaEnd.getGranule();
	    Granule m = deltaEnd.getGranule();
	    return i.compareTo(j) == 0 && k.compareTo(m) < 0;
	}

	  /** 
	   * Does this Period finish at the same time as another, but start after it,
	   * that is, is alpha.end == beta.end but alpha.start < beta.start?  
	   * @param alpha - Period to compare
	   * @param beta - Period to compare
	   * @return alpha.end == beta.end AND alpha.start < beta.start
	 * @throws TemporalDataTypeFormationException 
	   **/
    public boolean finishes(Period alpha, Period beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException{
	    // Determine granularity to convert alpha and/or beta to
    	Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());

		Period gamma = alpha.cast(toGran);
		Period delta = beta.cast(toGran);
	    Instant gammaStart = gamma.getStarting();
	    Instant gammaEnd = gamma.getEnding();
	    Instant deltaStart = delta.getStarting();
	    Instant deltaEnd = delta.getEnding();

		// Get both granules
		Granule i = gammaStart.getGranule();
		Granule j = deltaStart.getGranule();
		Granule k = gammaEnd.getGranule();
		Granule m = deltaEnd.getGranule();
		return i.compareTo(j) < 0 && k.compareTo(m) == 0;
    }	
		
   /** 
	* Does the end of a Period meet the start of another Period,
	* that is, is alpha.end == beta.start?  This is the inverse of metBy().
	* @param alpha - Period to compare
	* @param beta - Period to compare
	* @return alpha.end == beta.start
 * @throws TemporalDataTypeFormationException 
    **/
   public boolean meets(Period alpha, Period beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException{
	    // Determine granularity to convert alpha and/or beta to
	    Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());

	    Instant alphaEnd = alpha.getEnding();
	    Instant betaStart = beta.getStarting();
	    Instant gamma = alphaEnd.cast(toGran);
	    Instant delta = betaStart.cast(toGran);

	    // Get both granules
	    Granule i = gamma.getGranule();
	    Granule j = delta.getGranule();
	    return i.compareTo(j) == 0;
	}	
	
   /** 
    * Is the start of a Period metBy the start of another Period,
    * that is, is beta.end == alpha.start?  This is the inverse of meets().
    * @param alpha - Period to compare
    * @param beta - Period to compare
    * @return alpha.start == beta.end
 * @throws TemporalDataTypeFormationException 
    **/
   public boolean metBy(Period alpha, Period beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException{
	  // Determine granularity to convert alpha and/or beta to
	  Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());

      Instant alphaStart = alpha.getStarting();
      Instant betaEnd = beta.getEnding();
      Instant gamma = alphaStart.cast(toGran);
      Instant delta = betaEnd.cast(toGran);

      // Get both granules
      Granule i = gamma.getGranule();
      Granule j = delta.getGranule();
      return i.compareTo(j) == 0;
   }

   /** 
    * Does one Period overlap another?  Two Periods overlap if they intersect
    * in time.  This operation is symmetric.
    * @param alpha - Period to compare
    * @param beta - Period to compare
    * @return alpha overlaps beta
 * @throws TemporalDataTypeFormationException 
    **/
   public boolean overlaps(Period alpha, Period beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException{
	  // Determine granularity to convert alpha and/or beta to
	  Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());

		Period gamma = alpha.cast(toGran);
		Period delta = beta.cast(toGran);
	    Instant gammaStart = gamma.getStarting();
	    Instant gammaEnd = gamma.getEnding();
	    Instant deltaStart = delta.getStarting();
	    Instant deltaEnd = delta.getEnding();

      // Get both granules
      Granule i = gammaStart.getGranule();
      Granule j = deltaStart.getGranule();
      Granule k = gammaEnd.getGranule();
      Granule m = deltaEnd.getGranule();
      return !(k.compareTo(j) < 0 || m.compareTo(i) < 0);
   }


   /** 
    * Does this Period contain another?  A periods contains another if
    * it starts before it and ends after it.  The containment need not
    * be strict.
    * @param alpha - Period to compare
    * @param beta - Period to compare
    * @return alpha.start <= beta.start AND alpha.end >= beta.end
 * @throws TemporalDataTypeFormationException 
    **/
    public boolean contains(Period alpha, Period beta)throws CalendricSystemServiceException, TemporalDataTypeFormationException {
  	  // Determine granularity to convert alpha and/or beta to
      Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());

		Period gamma = alpha.cast(toGran);
		Period delta = beta.cast(toGran);
	    Instant gammaStart = gamma.getStarting();
	    Instant gammaEnd = gamma.getEnding();
	    Instant deltaStart = delta.getStarting();
	    Instant deltaEnd = delta.getEnding();

      // Get both granules
      Granule i = gammaStart.getGranule();
      Granule j = deltaStart.getGranule();
      Granule k = gammaEnd.getGranule();
      Granule m = deltaEnd.getGranule();
      return i.compareTo(j) <= 0 && k.compareTo(m) >= 0;
    }

    /** 
     * Is this Period during another?  A periods is during another if
     * it starts after it and ends before it.  The containment need not
     * be strict. 
     * @param alpha - Period to compare
     * @param beta - Period to compare
     * @return alpha.start >= beta.start AND alpha.end <= beta.end
     * @throws TemporalDataTypeFormationException 
     **/
    public boolean during(Period alpha, Period beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException{
       // Determine granularity to convert alpha and/or beta to
       Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());

		Period gamma = alpha.cast(toGran);
		Period delta = beta.cast(toGran);
	    Instant gammaStart = gamma.getStarting();
	    Instant gammaEnd = gamma.getEnding();
	    Instant deltaStart = delta.getStarting();
	    Instant deltaEnd = delta.getEnding();

       // Get both granules
       Granule i = gammaStart.getGranule();
       Granule j = deltaStart.getGranule();
       Granule k = gammaEnd.getGranule();
       Granule m = deltaEnd.getGranule();
       return i.compareTo(j) >= 0 && k.compareTo(m) <= 0;
    }    

    /** 
     * Does this Period represent the same times as another?  
     * A period is equalTo another if it both starts and finishes it.
     * This operation is symmetric.
     * @param alpha - Period to compare
     * @param beta - Period to compare
     * @return alpha.start == beta.start AND alpha.end == beta.end
     * @throws TemporalDataTypeFormationException 
     **/
    public boolean equalTo(Period alpha, Period beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException{
       // Determine granularity to convert alpha and/or beta to
       Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());

		Period gamma = alpha.cast(toGran);
		Period delta = beta.cast(toGran);
	    Instant gammaStart = gamma.getStarting();
	    Instant gammaEnd = gamma.getEnding();
	    Instant deltaStart = delta.getStarting();
	    Instant deltaEnd = delta.getEnding();

       // Get both granules
       Granule i = gammaStart.getGranule();
       Granule j = deltaStart.getGranule();
       Granule k = gammaEnd.getGranule();
       Granule m = deltaEnd.getGranule();
       return i.compareTo(j) == 0 && k.compareTo(m) == 0;
    }    

    /** 
     * Does this Instant represent the same times as another?  
     * An instant is equalTo another if it is at the same time.
     * This operation is symmetric.
     * @param alpha - Instant to compare
     * @param beta - Instant to compare
     * @return alpha == beta
     * @throws TemporalDataTypeFormationException 
     **/
    public boolean equalTo(Instant alpha, Instant beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
       // Determine granularity to convert alpha and/or beta to
       Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());


       Instant gamma = alpha.cast(toGran);
       Instant delta = beta.cast(toGran);
       // Get both granules
       Granule i = gamma.getGranule();
       Granule j = delta.getGranule();
       return i.compareTo(j) == 0;
    }    

    /** 
     * Displace (add) a Period on the time-line by an Interval 
     * @param alpha - displacee
     * @param beta - displacer
     * @return - [alpha.start + beta, alpha.end + beta]
     * @throws TemporalDataTypeFormationException 
     **/
    public Period add(Period alpha, Interval beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException{
       // Determine granularity to convert alpha and/or beta to
       Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());
       Period gamma = alpha.cast(toGran);
       Instant gammaEnd = gamma.getEnding();
       Instant gammaStart = gamma.getStarting();
       Interval delta = beta.cast(toGran);

       // Get both granules
       Granule i = gammaStart.getGranule();
       Granule j = delta.smallest();
       Granule k = gammaEnd.getGranule();
       Granule m = delta.largest();
       Granule low = i.add(j);
       Granule high = k.add(m);
       return new Period(toGran, 
                         new Instant(low), new Instant(high));
    }    

    /** 
     * Displace (add) an Instant on the time-line by an Interval 
     * @param alpha - displacee
     * @param beta - displacer
     * @return - alpha + beta
     * @throws TemporalDataTypeFormationException 
     **/
    public Instant add(Instant alpha, Interval beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException{
       // Determine granularity to convert alpha and/or beta to
       Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());

       Instant gamma = alpha.cast(toGran);
       Interval delta = beta.cast(toGran);
       // Get both granules
       Granule i = gamma.getGranule();
       Granule j = delta.smallest();
       Granule g = i.add(j);
       return new Instant(toGran, g);
    }
     
    /** 
     * Add one Interval to another
     * @param alpha - displacee
     * @param beta - displacer
     * @return - alpha + beta
     **/
     public Interval add(Interval alpha, Interval beta) throws CalendricSystemServiceException{
       // Determine granularity to convert alpha and/or beta to
       Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());

       Interval gamma = alpha.cast(toGran);
       Interval delta = beta.cast(toGran);
       // Get both granules
       Granule i = gamma.smallest();
       Granule j = delta.smallest();
       Granule g = i.add(j);
       return new Interval(toGran, g);
    }

     /** 
      * Displace (subtract) a Period on the time-line by an Interval 
      * @param alpha - displacee
      * @param beta - displacer
      * @return - [alpha.start - beta, alpha.end - beta]
     * @throws TemporalDataTypeFormationException 
      **/
     public Period subtract(Period alpha, Interval beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException{
        // Determine granularity to convert alpha and/or beta to
    	Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());

        Period gamma = alpha.cast(toGran);
        Instant gammaEnd = gamma.getEnding();
        Instant gammaStart = gamma.getStarting();
        Interval delta = beta.cast(toGran);

        // Get both granules
        Granule i = gammaStart.getGranule();
        Granule j = delta.smallest();
        Granule k = gammaEnd.getGranule();
        Granule m = delta.largest();
        Granule low = i.subtract(j);
        Granule high = k.subtract(m);
        return new Period(toGran, 
                           new Instant(low), new Instant(high));
     }     

     /** 
      * Displace (subtract) an Instant on the time-line by an Interval 
      * @param alpha - displacee
      * @param beta - displacer
      * @return - alpha - beta
     * @throws TemporalDataTypeFormationException 
      **/
     public Instant subtract(Instant alpha, Interval beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
        // Determine granularity to convert alpha and/or beta to
    	Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());

        Instant gamma = alpha.cast(toGran);
        Interval delta = beta.cast(toGran);
        // Get both granules
        Granule i = gamma.getGranule();
        Granule j = delta.smallest();
        Granule g = i.subtract(j);
        return new Instant(toGran, g);
     }    

     /** 
      * Subtract one Interval from another
      * @param alpha - displacee
      * @param beta - displacer
      * @return - alpha - beta
      **/
     public Interval subtract(Interval alpha, Interval beta) throws CalendricSystemServiceException{
        // Determine granularity to convert alpha and/or beta to
    	Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());

        Interval gamma = alpha.cast(toGran);
        Interval delta = beta.cast(toGran);
        // Get both granules
        Granule i = gamma.smallest();
        Granule j = delta.smallest();
        Granule g = i.subtract(j);
        return new Interval(toGran, g);
     }
     
     /** 
      * Negate an Interval. 
      * @param alpha - Interval to negate
      * @return - -alpha
      **/
      public Interval negate(Interval alpha) {
        // Get both granules
        Granule i = alpha.smallest();
        return new Interval(alpha.getGranularity(), i.negate());
      }    

     /** 
      * Multiply an Interval by an integer constant.  
      * @param alpha - Interval to multiply
      * @param n - by how many 
      * @return result - (alpha * n)
      **/
     public Interval multiply(Interval alpha, int n) {
        // Get both granules
        Granule i = alpha.smallest();
        return new Interval(alpha.getGranularity(), i.multiply(n));
     }

     /** 
      * Divide an Interval by an integer constant.  
      * @param alpha - Interval to divide
      * @param n - by how many 
      * @return result - (alpha div n)
      **/
     public Interval divide(Interval alpha, int n) {
        // Get both granules
        Granule i = alpha.smallest();
        return new Interval(alpha.getGranularity(), i.divide(n));
     }

     /** 
      * Divide an Interval by an Interval (integer division). 
      * @param alpha - divisee
      * @param beta - divisor
      * @param result - alpha div beta
      * @status NOT IMPLEMENTED - problematic for MAX/MIN
      **/
     public int divide(Interval alpha, Interval beta) throws CalendricSystemServiceException{
        return 0;
     }  

     /** 
     * Temporal constructor - Choose the first (earliest) from among a pair of instants
     * @param alpha - Instant 
     * @param beta - Instant 
     * @return first Instant on the time-line from among alpha and beta
     * @throws TemporalDataTypeFormationException 
     **/
     public Instant first(Instant alpha, Instant beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException{
        // Determine granularity to convert alpha and/or beta to
    	Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());

        Instant gamma = alpha.cast(toGran);
        Instant delta = beta.cast(toGran);
        Granule i = gamma.getGranule();
        Granule j = delta.getGranule();
        if (i.compareTo(j) < 0) { return gamma; }
        return delta;
     }

     /** 
      * Temporal constructor - Choose the last (latest) from among a pair of instants
      * @param alpha - Instant 
      * @param beta - Instant 
      * @return last Instant on the time-line from among alpha and beta
     * @throws TemporalDataTypeFormationException 
      **/
     public Instant last(Instant alpha, Instant beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException{
        // Determine granularity to convert alpha and/or beta to
    	Granularity toGran = lattice.greatestLowerBound(alpha.getGranularity(), beta.getGranularity());

        Instant gamma = alpha.cast(toGran);
        Instant delta = beta.cast(toGran);
        Granule i = gamma.getGranule();
        Granule j = delta.getGranule();
        if (i.compareTo(j) < 0) { return delta; }
        return gamma;
     }

}
