package tauzaman.temporaldatatypes.semantics;

import tauzaman.calendricsystem.CalendricSystemServiceException;
import tauzaman.calendricsystem.granularitylattice.GranularityLattice;
import tauzaman.extendedboolean.ExtendedBoolean;
import tauzaman.timestamp.*;
import tauzaman.temporaldatatypes.*;

/**
* The <code>IndeterminateSemantics</code> interface contains all the 
* arithmetic and comparison operations that are available on 
* indeterminate time values. 
* In an indeterminate semantics, granularity conversions are scales, 
* and <code>ExtendedBoolean</code>'s are the logic values. The comparison 
* operations are also given a plausibility, which is an integer representing 
* the percent threshold which the comparison's probability must exceed to return
* True. If a plausibility is not given, the current plausibility of the 
* <code>IndeterminateSemantics</code> is used. The plausibility of a 
* <code>IndeterminateSemantics</code> can be set, the default is <code>DEFINITE</code>.
* This is an interface because there are lots of different semantics
* for even standard operations such as precedes, such as left-operand
* semantics where the operation is performed at the granularity of the left 
* operand.
* Below is an example of how one might use a semantics (assume that
* LeftOperandIndeterminateSemantics implements IndeterminateSemantics).
*  <pre>
*     LeftOperandIndeterminateSemantics ops = new LeftOperandIndeterminateSemantics();
*     Instant j = new Instant("July 3, 2003");
*     Interval i = new Interval("1 month");
*     Instant k = ops.add(i,j); // k is in the granularity of "days"
*     Instant m = ops.add(j,i); // k is in the granularity of "months"
*  </pre>
* Just a few operations are currently represented to give an idea of the
* eventual interface.  All the operations are on individual kinds of
* time values, e.g., Instant, Period, Interval.   We need to expand this
* with a complete set of temporal operators implementing Allen's algebra.
*
* @author  Curtis Dyreson, Bedirhan Urgun, Alex Henniges, and Richard Snodgrass
* @version 2.0, Sep/2/2008
* @see     tauzaman.temporaldatatypes.Instant
* @see     tauzaman.temporaldatatypes.Period
* @see     tauzaman.temporaldatatypes.Interval
**/
public abstract class IndeterminateSemantics {
  /**
   * <code>DEFINITE</code> denotes a comparison that returns True only 
   * if the comparison is certain to be True.
   */
  public static final int DEFINITE = 100;
  /**
   * <code>POSSIBLE</code> denotes a comparison that returns True as 
   * long as there is some possibility that the comparison is True.
   */
  public static final int POSSIBLE = 0;
  
  public enum Credibility {
	  Indeterminate, Min, Max, Expected
  };
  
  protected int plausibility = DEFINITE;
  protected GranularityLattice lattice;
  protected Instant leftInstant;
  protected Instant rightInstant;
  protected Period leftPeriod;
  protected Period rightPeriod;
  protected Interval leftInterval;
  protected Interval rightInterval;
  
  /**
   * Sets this <code>IndeterminateSemantics</code> plausibility to the
   * given value. The set plausibility is used in all comparisons where 
   * a plausibility is not explicitly defined.
   * 
   * @param plausibility
   * 		The percent of plausibility to set for this 
   * 		<code>IndeterminateSemantics</code>. The set of valid values are
   * 		<I>{POSSIBLE, 1, ..., 99, DEFINITE}</I>.
   * @throws IllegalArgumentException
   * 		If the given plausibility is not within the range of valid values.
   */
  public void setPlausibility(int plausibility) throws IllegalArgumentException {
		if(plausibility < 0 || plausibility > 100) {
			throw new IllegalArgumentException("" + plausibility + " is an invalid value for plausibility");
		}
	  this.plausibility = plausibility;
  }
  
  public void setLattice(GranularityLattice lattice) {
	  this.lattice = lattice;
  }
  
  public abstract void scalePair(Instant left, Instant right) throws CalendricSystemServiceException;
  public abstract void scalePair(Period left, Period right) throws CalendricSystemServiceException;
  public abstract void scalePair(Period left, Interval right) throws CalendricSystemServiceException;
  public abstract void scalePair(Interval left, Interval right) throws CalendricSystemServiceException;
  public abstract void scalePair(Instant left, Interval right) throws CalendricSystemServiceException;
  
  /** 
  * Does one Instant precede another, that is, is alpha < beta by a preset plausibility?
  * @param alpha  Instant to compare
  * @param beta  Instant to compare
  * @return <code>ExtendedBoolean</code> for alpha < beta.
  * @throws TemporalDataTypeFormationException
  * 	If it cannot create the scaled time value.  
  * @throws CalendricSystemServiceException
  * 	If it cannot convert granularities.
  **/
  public ExtendedBoolean precedes(Instant alpha, Instant beta) 
  		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	  return precedes(alpha, beta, this.plausibility);
  }
  
  /** 
   * Does one Instant precede another, that is, is alpha < beta by a given plausibility?
   * @param alpha  Instant to compare
   * @param beta  Instant to compare
   * @param plausibility The probability threshold for alpha < beta
   * @return <code>ExtendedBoolean</code> for alpha < beta.
   * @throws TemporalDataTypeFormationException
   * 	If it cannot create the scaled time value.  
   * @throws CalendricSystemServiceException
   * 	If it cannot convert granularities.
   **/
   public ExtendedBoolean precedes(Instant alpha, Instant beta, int plausibility) 
   		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	    scalePair(alpha, beta);
		Granule i = leftInstant.getGranule();
		Granule j = rightInstant.getGranule();
		return i.lessThan(j, plausibility);
   }

  /** 
  * Does an Instant precede the start of a Period, 
  * that is, is alpha < beta.start by a preset plausibility?
  * @param alpha  Instant to compare
  * @param beta  Period to compare
  * @return <code>ExtendedBoolean</code> for alpha < beta.start
  * @throws TemporalDataTypeFormationException
  * 	If it cannot create the scaled time value.  
  * @throws CalendricSystemServiceException
  * 	If it cannot convert granularities.
  **/
  public ExtendedBoolean precedes(Instant alpha, Period beta) 
  		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	  
	  return precedes(alpha, beta, this.plausibility);
  }

  /** 
   * Does an Instant precede the start of a Period, 
   * that is, is alpha < beta.start by a given plausibility?
   * @param alpha  Instant to compare
   * @param beta  Period to compare
   * @param plausibility The probability threshold for alpha < beta.start
   * @return <code>ExtendedBoolean</code> for alpha < beta.start
   * @throws TemporalDataTypeFormationException
   * 	If it cannot create the scaled time value.  
   * @throws CalendricSystemServiceException
   * 	If it cannot convert granularities. 
   **/
   public ExtendedBoolean precedes(Instant alpha, Period beta, int plausibility) 
   			throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	     scalePair(alpha, beta.getStarting());
		 Granule i = leftInstant.getGranule();
		 Granule j = rightInstant.getGranule(); 
		 return i.lessThan(j, plausibility);
   }
  
  /** 
  * Does the end of a Period precede the start of an Instant, 
  * that is, is alpha.end < beta by a preset plausibility?
  * @param alpha  Period to compare
  * @param beta  Instant to compare
  * @return <code>ExtendedBoolean</code> for alpha.end < beta
  * @throws TemporalDataTypeFormationException
  * 	If it cannot create the scaled time value.  
  * @throws CalendricSystemServiceException
  * 	If it cannot convert granularities. 
  **/
  public ExtendedBoolean precedes(Period alpha, Instant beta) 
  		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	  return precedes(alpha, beta, this.plausibility);
  }
  
  /** 
   * Does the end of a Period precede the start of an Instant, 
   * that is, is alpha.end < beta by a given plausibility?
   * @param alpha  Period to compare
   * @param beta  Instant to compare
   * @param plausibility The probability threshold for alpha.end < beta
   * @return <code>ExtendedBoolean</code> for alpha.end < beta
   * @throws TemporalDataTypeFormationException
   * 	If it cannot create the scaled time value.  
   * @throws CalendricSystemServiceException
   * 	If it cannot convert granularities.  
   **/
   public ExtendedBoolean precedes(Period alpha, Instant beta, int plausibility) 
   		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	    scalePair(alpha.getEnding(), beta);
		Granule i = leftInstant.getGranule();
		Granule j = rightInstant.getGranule();
	    return i.lessThan(j, plausibility);
   }

  /** 
  * Does the end of a Period precede the start of another,
  * that is, is alpha.end < beta.start by a preset plausibility?
  * @param alpha  Period to compare
  * @param beta  Period to compare
  * @return <code>ExtendedBoolean</code> for alpha.end < beta.start
  * @throws TemporalDataTypeFormationException
  * 	If it cannot create the scaled time value.  
  * @throws CalendricSystemServiceException
  * 	If it cannot convert granularities. 
  **/
  public ExtendedBoolean precedes(Period alpha, Period beta) 
  		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	  return precedes(alpha, beta, this.plausibility);
  }
  
  /** 
   * Does the end of a Period precede the start of another,
   * that is, is alpha.end < beta.start by a given plausibility?
   * @param alpha  Period to compare
   * @param beta  Period to compare
   * @param plausibility The probability threshold for alpha.end < beta.start
   * @return <code>ExtendedBoolean</code> for alpha.end < beta.start
   * @throws TemporalDataTypeFormationException
   * 	If it cannot create the scaled time value.  
   * @throws CalendricSystemServiceException
   * 	If it cannot convert granularities. 
   **/
   public ExtendedBoolean precedes(Period alpha, Period beta, int plausibility) 
   		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	    scalePair(alpha.getEnding(), beta.getStarting());
		Granule i = leftInstant.getGranule();
		Granule j = rightInstant.getGranule();
	    return i.lessThan(j, plausibility);
   }

  /** 
  * Does this Period begin at the same time as another, but end before it,
  * that is, is alpha.start == beta.start but alpha.end < beta.end by a preset plausibility?  
  * @param alpha  Period to compare
  * @param beta  Period to compare
  * @return <code>ExtendedBoolean</code> for alpha.start == beta.start AND alpha.end < beta.end
  * @throws TemporalDataTypeFormationException
  * 	If it cannot create the scaled time value.  
  * @throws CalendricSystemServiceException
  * 	If it cannot convert granularities.
  **/
  public ExtendedBoolean starts(Period alpha, Period beta) 
  		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	  return starts(alpha, beta, this.plausibility);
  }
  
  /** 
   * Does this Period begin at the same time as another, but end before it,
   * that is, is alpha.start == beta.start but alpha.end < beta.end  by a given plausibility?  
   * @param alpha  Period to compare
   * @param beta  Period to compare
   * @param plausibility The probability threshold for alpha.start == beta.start AND alpha.end < beta.end
   * @return <code>ExtendedBoolean</code> for alpha.start == beta.start AND alpha.end < beta.end
   * @throws TemporalDataTypeFormationException
   * 	If it cannot create the scaled time value.  
   * @throws CalendricSystemServiceException
   * 	If it cannot convert granularities.  
   **/
   public ExtendedBoolean starts(Period alpha, Period beta, int plausibility) 
   		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	    scalePair(alpha, beta);
	    Granule i = leftPeriod.getStarting().getGranule();
	    Granule j = rightPeriod.getStarting().getGranule();
	    Granule k = leftPeriod.getEnding().getGranule();
	    Granule m = rightPeriod.getEnding().getGranule();
	    ExtendedBoolean b1 = i.equalTo(j, plausibility);
	    ExtendedBoolean b2 = k.lessThan(m, plausibility);
	    return b1.and(b2);
   }

  /** 
  * Does this Period finish at the same time as another, but start after it,
  * that is, is alpha.end == beta.end but alpha.start < beta.start by a preset plausibility?  
  * @param alpha  Period to compare
  * @param beta  Period to compare
  * @return <code>ExtendedBoolean</code> for alpha.end == beta.end AND alpha.start < beta.start
  * @throws TemporalDataTypeFormationException
  * 	If it cannot create the scaled time value.  
  * @throws CalendricSystemServiceException
  * 	If it cannot convert granularities.
  **/
  public ExtendedBoolean finishes(Period alpha, Period beta) 
  		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	  return finishes(alpha, beta, this.plausibility);
  }
  
  /** 
   * Does this Period finish at the same time as another, but start after it,
   * that is, is alpha.end == beta.end but alpha.start < beta.start by a given plausibility?  
   * @param alpha  Period to compare
   * @param beta  Period to compare
   * @param plausibility The probability threshold for alpha.end == beta.end AND alpha.start < beta.start
   * @return <code>ExtendedBoolean</code> for alpha.end == beta.end AND alpha.start < beta.start
   * @throws TemporalDataTypeFormationException
   * 	If it cannot create the scaled time value.  
   * @throws CalendricSystemServiceException
   * 	If it cannot convert granularities. 
   **/
   public ExtendedBoolean finishes(Period alpha, Period beta, int plausibility) 
   		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	    scalePair(alpha, beta);
	    Granule i = leftPeriod.getStarting().getGranule();
	    Granule j = rightPeriod.getStarting().getGranule();
	    Granule k = leftPeriod.getEnding().getGranule();
	    Granule m = rightPeriod.getEnding().getGranule();
	    ExtendedBoolean b1 = i.lessThan(j, plausibility);
	    ExtendedBoolean b2 = k.equalTo(m, plausibility); 
	    return b1.and(b2);
   }

  /** 
  * Does the end of a Period meet the start of another Period,
  * that is, is alpha.end == beta.start by a preset plausibility?  This is the inverse of metBy().
  * @param alpha  Period to compare
  * @param beta  Period to compare
  * @return <code>ExtendedBoolean</code> for alpha.end == beta.start
  * @throws TemporalDataTypeFormationException
  * 	If it cannot create the scaled time value.  
  * @throws CalendricSystemServiceException
  * 	If it cannot convert granularities.
  **/
  public ExtendedBoolean meets(Period alpha, Period beta) 
  		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	  return meets(alpha, beta, this.plausibility);
  }
  
  /** 
   * Does the end of a Period meet the start of another Period,
   * that is, is alpha.end == beta.start by a given plausibility?  This is the inverse of metBy().
   * @param alpha  Period to compare
   * @param beta  Period to compare
   * @param plausibility The probability threshold for alpha.end == beta.start
   * @return <code>ExtendedBoolean</code> for alpha.end == beta.start
   * @throws TemporalDataTypeFormationException
   * 	If it cannot create the scaled time value.  
   * @throws CalendricSystemServiceException
   * 	If it cannot convert granularities.
   **/
   public ExtendedBoolean meets(Period alpha, Period beta, int plausibility) 
   		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	    scalePair(alpha.getEnding(), beta.getStarting());
	    Granule i = leftInstant.getGranule();
	    Granule j = rightInstant.getGranule();
	    return i.equalTo(j, plausibility);
   }

  /** 
  * Is the start of a Period metBy the start of another Period,
  * that is, is beta.end == alpha.start by a preset plausibility?  This is the inverse of meets().
  * @param alpha  Period to compare
  * @param beta  Period to compare
  * @return <code>ExtendedBoolean</code> for alpha.start == beta.end
  * @throws TemporalDataTypeFormationException
  * 	If it cannot create the scaled time value.  
  * @throws CalendricSystemServiceException
  * 	If it cannot convert granularities.
  **/
  public ExtendedBoolean metBy(Period alpha, Period beta) 
  		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	  return metBy(alpha, beta, this.plausibility);
  }
  
  /** 
   * Is the start of a Period metBy the start of another Period,
   * that is, is beta.end == alpha.start by a given plausibility?  This is the inverse of meets().
   * @param alpha  Period to compare
   * @param beta  Period to compare
   * @param plausibility The probability threshold for alpha.start == beta.end
   * @return <code>ExtendedBoolean</code> for alpha.start == beta.end
   * @throws TemporalDataTypeFormationException
   * 	If it cannot create the scaled time value.  
   * @throws CalendricSystemServiceException
   * 	If it cannot convert granularities. 
   **/
   public ExtendedBoolean metBy(Period alpha, Period beta, int plausibility) 
   		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	    scalePair(alpha.getStarting(), beta.getEnding());
	    Granule i = leftInstant.getGranule();
        Granule j = rightInstant.getGranule();
	    return i.equalTo(j, plausibility);
   }

  /** 
  * Does one Period overlap another by a preset plausibility?  Two Periods overlap if they intersect
  * in time.  This operation is symmetric.
  * @param alpha  Period to compare
  * @param beta  Period to compare
  * @return <code>ExtendedBoolean</code> for alpha overlaps beta
  * @throws TemporalDataTypeFormationException
  * 	If it cannot create the scaled time value.  
  * @throws CalendricSystemServiceException
  * 	If it cannot convert granularities.
  **/
  public ExtendedBoolean overlaps(Period alpha, Period beta) 
  		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	  return overlaps(alpha, beta, this.plausibility);
  }
  
  /** 
   * Does one Period overlap another by a given plausibility?  Two Periods overlap if they intersect
   * in time.  This operation is symmetric.
   * @param alpha  Period to compare
   * @param beta  Period to compare
   * @param plausibility The probability threshold for alpha overlaps beta
   * @return <code>ExtendedBoolean</code> for alpha overlaps beta
   * @throws TemporalDataTypeFormationException
   * 	If it cannot create the scaled time value.  
   * @throws CalendricSystemServiceException
   * 	If it cannot convert granularities.
   **/
   public ExtendedBoolean overlaps(Period alpha, Period beta, int plausibility) 
   		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	    scalePair(alpha, beta);
	    Granule i = leftPeriod.getStarting().getGranule();
	    Granule j = rightPeriod.getStarting().getGranule();
	    Granule k = leftPeriod.getEnding().getGranule();
	    Granule m = rightPeriod.getEnding().getGranule();
	    ExtendedBoolean b1 = k.lessThan(j, plausibility);
	    ExtendedBoolean b2 = m.lessThan(i, plausibility);
	    ExtendedBoolean b3 = b1.or(b2);
	    return b3.not();
   }

  /** 
  * Does this Period contain another by a preset plausibility?  A periods contains another if
  * it starts before it and ends after it.  The containment need not
  * be strict.
  * @param alpha  Period to compare
  * @param beta  Period to compare
  * @return <code>ExtendedBoolean</code> for alpha.start <= beta.start AND alpha.end >= beta.end
  * @throws TemporalDataTypeFormationException
  * 	If it cannot create the scaled time value.  
  * @throws CalendricSystemServiceException
  * 	If it cannot convert granularities.
  **/
  public ExtendedBoolean contains(Period alpha, Period beta) 
  		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	  return contains(alpha, beta, this.plausibility);
  }
  
  /** 
   * Does this Period contain another by a given plausibility?  A periods contains another if
   * it starts before it and ends after it.  The containment need not
   * be strict.
   * @param alpha  Period to compare
   * @param beta  Period to compare
   * @param plausibility The probability threshold for alpha.start <= beta.start AND alpha.end >= beta.end
   * @return <code>ExtendedBoolean</code> for alpha.start <= beta.start AND alpha.end >= beta.end
   * @throws TemporalDataTypeFormationException
   * 	If it cannot create the scaled time value.  
   * @throws CalendricSystemServiceException
   * 	If it cannot convert granularities.
   **/
   public ExtendedBoolean contains(Period alpha, Period beta, int plausibility) 
   		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	    scalePair(alpha, beta);
	    Granule i = leftPeriod.getStarting().getGranule();
	    Granule j = rightPeriod.getStarting().getGranule();
	    Granule k = leftPeriod.getEnding().getGranule();
	    Granule m = rightPeriod.getEnding().getGranule();
	    ExtendedBoolean b1 = i.lessThanOrEqualTo(j, plausibility);
	    ExtendedBoolean b2 = m.lessThanOrEqualTo(k, plausibility);
	    return b1.and(b2);
   }

  /** 
  * Is this Period during another by a preset plausibility?  A periods is during another if
  * it starts after it and ends before it.  The containment need not
  * be strict. 
  * @param alpha  Period to compare
  * @param beta  Period to compare
  * @return <code>ExtendedBoolean</code> for alpha.start >= beta.start AND alpha.end <= beta.end
  * @throws TemporalDataTypeFormationException
  * 	If it cannot create the scaled time value.  
  * @throws CalendricSystemServiceException
  * 	If it cannot convert granularities.
  **/
  public ExtendedBoolean during(Period alpha, Period beta) 
  		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	  return during(alpha, beta, this.plausibility);
  }
  
  /** 
   * Is this Period during another by a given plausibility?  A periods is during another if
   * it starts after it and ends before it.  The containment need not
   * be strict. 
   * @param alpha  Period to compare
   * @param beta  Period to compare
   * @param plausibility The probability threshold for alpha.start >= beta.start AND alpha.end <= beta.end
   * @return <code>ExtendedBoolean</code> for alpha.start >= beta.start AND alpha.end <= beta.end
   * @throws TemporalDataTypeFormationException
   * 	If it cannot create the scaled time value.  
   * @throws CalendricSystemServiceException
   * 	If it cannot convert granularities.
   **/
   public ExtendedBoolean during(Period alpha, Period beta, int plausibility) 
   		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	    scalePair(alpha, beta);
	    Granule i = leftPeriod.getStarting().getGranule();
	    Granule j = rightPeriod.getStarting().getGranule();
	    Granule k = leftPeriod.getEnding().getGranule();
	    Granule m = rightPeriod.getEnding().getGranule();
	    ExtendedBoolean b1 = j.lessThanOrEqualTo(i, plausibility);
	    ExtendedBoolean b2 = k.lessThanOrEqualTo(m, plausibility); 
	    return b1.and(b2);
   }

  /** 
  * Does this Period represent the same times as another by a preset plausibility?  
  * A period is equalTo another if it both starts and finishes it.
  * This operation is symmetric.
  * @param alpha  Period to compare
  * @param beta  Period to compare
  * @return <code>ExtendedBoolean</code> for alpha.start == beta.start AND alpha.end == beta.end
  * @throws TemporalDataTypeFormationException
  * 	If it cannot create the scaled time value.  
  * @throws CalendricSystemServiceException
  * 	If it cannot convert granularities.
  **/
  public ExtendedBoolean equalTo(Period alpha, Period beta) 
  		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	  return equalTo(alpha, beta, this.plausibility);
  }
  
  /** 
   * Does this Period represent the same times as another by a given plausibility?  
   * A period is equalTo another if it both starts and finishes it.
   * This operation is symmetric.
   * @param alpha  Period to compare
   * @param beta  Period to compare
   * @param plausibility The probability threshold for alpha.start == beta.start AND alpha.end == beta.end
   * @return <code>ExtendedBoolean</code> for alpha.start == beta.start AND alpha.end == beta.end
   * @throws TemporalDataTypeFormationException
   * 	If it cannot create the scaled time value.  
   * @throws CalendricSystemServiceException
   * 	If it cannot convert granularities. 
   **/
   public ExtendedBoolean equalTo(Period alpha, Period beta, int plausibility) 
   		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	    scalePair(alpha, beta);
	    Granule i = leftPeriod.getStarting().getGranule();
	    Granule j = rightPeriod.getStarting().getGranule();
	    Granule k = leftPeriod.getEnding().getGranule();
	    Granule m = rightPeriod.getEnding().getGranule();
	    ExtendedBoolean b1 = i.equalTo(j, plausibility);
	    ExtendedBoolean b2 = k.equalTo(m, plausibility);
	    return b1.and(b2);
   }

  /** 
  * Does this Instant represent the same times as another by a preset plausibility?  
  * An instant is equalTo another if it is at the same time.
  * This operation is symmetric.
  * @param alpha  Instant to compare
  * @param beta  Instant to compare
  * @return <code>ExtendedBoolean</code> for alpha == beta
  * @throws TemporalDataTypeFormationException
  * 	If it cannot create the scaled time value.  
  * @throws CalendricSystemServiceException
  * 	If it cannot convert granularities.
  **/
  public ExtendedBoolean equalTo(Instant alpha, Instant beta) 
  		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	  return equalTo(alpha, beta, this.plausibility);
  }
  
  /** 
   * Does this Instant represent the same times as another by a given plausibility?  
   * An instant is equalTo another if it is at the same time.
   * This operation is symmetric.
   * @param alpha  Instant to compare
   * @param beta  Instant to compare
   * @param plausibility The probability threshold for alpha == beta
   * @return <code>ExtendedBoolean</code> for alpha == beta
   * @throws TemporalDataTypeFormationException
   * 	If it cannot create the scaled time value.  
   * @throws CalendricSystemServiceException
   * 	If it cannot convert granularities. 
   **/
   public ExtendedBoolean equalTo(Instant alpha, Instant beta, int plausibility) 
   		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	   scalePair(alpha, beta);
       Granule i = leftInstant.getGranule();
       Granule j = rightInstant.getGranule();
	   return i.equalTo(j, plausibility);
   }

	/** 
	 * Displace (add) a Period on the time-line by an Interval 
	 * @param alpha - displacee
	 * @param beta - displacer
	 * @return - [alpha.start + beta, alpha.end + beta]
	 * @throws TemporalDataTypeFormationException 
	 * @throws CalendricSystemServiceException 
	 **/
	public Period add(Period alpha, Interval beta) 
	throws TemporalDataTypeFormationException, CalendricSystemServiceException {
	       
		   scalePair(alpha, beta);

	       Granule i = leftPeriod.getStarting().getGranule();
	       Granule j = rightInterval.smallest();
	       Granule k = leftPeriod.getEnding().getGranule();
	       Granule m = rightInterval.largest();
	       Granule low = i.add(j);
	       Granule high = k.add(m);
	       return new Period(leftPeriod.getGranularity(), 
	                         new Instant(low), new Instant(high));
	}
	  
	/** 
	 * Displace (add) an Instant on the time-line by an Interval 
	 * @param alpha - displacee
	 * @param beta - displacer
	 * @return - alpha + beta
	 * @throws CalendricSystemServiceException 
	 * @throws TemporalDataTypeFormationException 
	 **/
	public Instant add(Instant alpha, Interval beta) 
	throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	    scalePair(alpha, beta);

	     // Get both granules
	     Granule i = leftInstant.getGranule();
	     Granule j = rightInterval.smallest();
	     Granule g = i.add(j);
	     return new Instant(leftInstant.getGranularity(), g);
	}
	  /** 
	  * Add one Interval to another
	  * @param alpha - displacee
	  * @param beta - displacer
	  * @return - alpha + beta
	 * @throws CalendricSystemServiceException 
	  **/
	  public Interval add(Interval alpha, Interval beta) throws CalendricSystemServiceException {
	       scalePair(alpha, beta);

	       // Get both granules
	       Granule i = leftInterval.smallest();
	       Granule j = rightInterval.smallest();
	       Granule g = i.add(j);
	       return new Interval(leftInterval.getGranularity(), g);
	  }

	  /** 
	   * Displace (subtract) a Period on the time-line by an Interval 
	   * @param alpha - displacee
	   * @param beta - displacer
	   * @return - [alpha.start - beta, alpha.end - beta]
	   * @throws CalendricSystemServiceException 
	   * @throws TemporalDataTypeFormationException 
	   **/
	  public Period subtract(Period alpha, Interval beta) 
	  throws CalendricSystemServiceException, TemporalDataTypeFormationException {
		  scalePair(alpha, beta);

	      Granule i = leftPeriod.getStarting().getGranule();
	      Granule j = rightInterval.smallest();
	      Granule k = leftPeriod.getEnding().getGranule();
	      Granule m = rightInterval.largest();
	      Granule low = i.subtract(j);
	      Granule high = k.subtract(m);
	      return new Period(leftPeriod.getGranularity(), 
	                     new Instant(low), new Instant(high));
	 }

	  /** 
	   * Displace (subtract) an Instant on the time-line by an Interval 
	   * @param alpha - displacee
	   * @param beta - displacer
	   * @return - alpha - beta
	   * @throws CalendricSystemServiceException 
	   * @throws TemporalDataTypeFormationException 
	   **/
	   public Instant subtract(Instant alpha, Interval beta) 
	   throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	       scalePair(alpha, beta);

	       Granule i = leftInstant.getGranule();
	       Granule j = rightInterval.smallest();
	       Granule g = i.subtract(j);
	       return new Instant(leftInstant.getGranularity(), g);
	  }

	  /** 
	   * Subtract one Interval from another
	   * @param alpha - displacee
	   * @param beta - displacer
	   * @return - alpha - beta
	   * @throws CalendricSystemServiceException 
	   **/
	   public Interval subtract(Interval alpha, Interval beta) throws CalendricSystemServiceException {
	       scalePair(alpha, beta);
		   
	       Granule i = leftInterval.smallest();
	       Granule j = rightInterval.smallest();
	       Granule g = i.subtract(j);
	       return new Interval(leftInterval.getGranularity(), g);
	  }
  /** 
  * Negate an Interval. 
  * @param alpha - Interval to negate
  * @return - -alpha
  **/
  public Interval negate(Interval alpha) {
	    Granule i = alpha.smallest();
	    return new Interval(alpha.getGranularity(), i.negate());
  }

  /** 
  * Multiply an Interval by an integer constant.  
  * @param alpha - Interval to multiply
  * @param n - by how many 
  * @return result - (alpha * n)
  **/
  public Interval multiply(Interval alpha, int n) {
	    Granule i = alpha.smallest();
	    return new Interval(alpha.getGranularity(), i.multiply(n));
  }

  /** 
  * Divide an Interval by an integer constant.  
  * @param alpha - Interval to divide
  * @param n - by how many 
  * @return result - (alpha div n)
  **/
  public Interval divide(Interval alpha, int n) {
	    Granule i = alpha.smallest();
	    return new Interval(alpha.getGranularity(), i.divide(n));
  }

  /** 
  * Divide an Interval by an Interval (integer division). 
  * @param alpha - divisee
  * @param beta - divisor
  * @param result - alpha div beta
  * @status NOT IMPLEMENTED - problematic for MAX/MIN
  **/
  public int divide(Interval alpha, Interval beta) {
	  return 0;
  }

  /** 
  * Temporal constructor - Choose the first (earliest) from among a pair of instants
  * @param alpha - Instant 
  * @param beta - Instant 
  * @return first Instant on the time-line from among alpha and beta
 * @throws TemporalDataTypeFormationException 
 * @throws CalendricSystemServiceException 
  **/
  public Instant first(Instant alpha, Instant beta) 
  		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	  scalePair(alpha, beta);
      Granule i = leftInstant.getGranule();
      Granule j = rightInstant.getGranule();
      if (i.compareTo(j) < 0) { return leftInstant; }
      return rightInstant;
  }

  /** 
  * Temporal constructor - Choose the last (latest) from among a pair of instants
  * @param alpha - Instant 
  * @param beta - Instant 
  * @return last Instant on the time-line from among alpha and beta
 * @throws TemporalDataTypeFormationException 
 * @throws CalendricSystemServiceException 
  **/
  public Instant last(Instant alpha, Instant beta) 
  		throws CalendricSystemServiceException, TemporalDataTypeFormationException {
	  scalePair(alpha, beta);
      Granule i = leftInstant.getGranule();
      Granule j = rightInstant.getGranule();
      if (i.compareTo(j) < 0) { return rightInstant; }
      return leftInstant;
  }
  
  public Instant narrow(Instant alpha, Credibility cred) throws TemporalDataTypeFormationException {
	  switch(cred) {
	  	case Indeterminate: return alpha;
	  	case Min: return new Instant(alpha.getGranule().narrow(cred));
	  	case Max: return new Instant(alpha.getGranule().narrow(cred));
	  	case Expected: return new Instant(alpha.getGranule().narrow(cred));
	  	default: return alpha;
	  }
  }
  
  public Period narrow(Period alpha, Credibility cred) throws TemporalDataTypeFormationException {
	  switch(cred) {
	  case Indeterminate: return alpha;
	  case Min: return new Period(narrow(alpha.getStarting(), Credibility.Min), narrow(alpha.getEnding(), Credibility.Max));
	  case Max: return new Period(narrow(alpha.getStarting(), Credibility.Max), narrow(alpha.getEnding(), Credibility.Min));
	  case Expected: return new Period(narrow(alpha.getStarting(), Credibility.Expected), narrow(alpha.getEnding(), Credibility.Expected));
	  default: return alpha;
	  }
  }
}
