package tauzaman.temporaldatatypes.semantics;

import tauzaman.calendricsystem.CalendricSystemServiceException;
import tauzaman.temporaldatatypes.Instant;
import tauzaman.temporaldatatypes.Interval;
import tauzaman.temporaldatatypes.Period;
import tauzaman.temporaldatatypes.TemporalDataTypeFormationException;
import tauzaman.timestamp.*;

/**
* The <code>LeftOperandDeterminateSemantics</code> class implements
* the <code>DeterminateSemantics</code> interface to convert the
* right operand to the granularity of the left operand for each
* operation.
*
* Below is an example of how one might use this semantics.
*  <pre>
*     LeftOperandDeterminateSemantics ops = new LeftOperandDeterminateSemantics();
*     Instant j = new Instant("July 3, 2003");
*     Interval i = new Interval("1 month");
*     Instant k = ops.add(i,j); // k is in the granularity of "days"
*     Instant m = ops.add(j,i); // k is in the granularity of "months"
*  </pre>
* Just a few operations are currently represented to give an idea of the
* eventual interface.  All the operations are on individual kinds of
* time values, e.g., Instant, Period, Interval.   We need to expand this
* with a complete set of temporal operators implementing Allen's algebra.
*
* @author  Curtis Dyreson and Bedirhan Urgun
* @version 0.9, Jan/31/2003
* @see     tauzaman.temporaldatatypes.Instant
* @see     tauzaman.temporaldatatypes.Period
* @see     tauzaman.temporaldatatypes.Interval
**/
public class LeftOperandDeterminateSemantics implements DeterminateSemantics {
  /** 
  * Does one Instant precede another, that is, is alpha < beta?
  * @param alpha - Instant to compare
  * @param beta - Instant to compare
  * @return alpha < beta
 * @throws TemporalDataTypeFormationException 
 * @throws CalendricSystemServiceException 
  **/
  public boolean precedes(Instant alpha, Instant beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
    // Convert beta to alpha's granularity
    Instant gamma = beta.cast(alpha.getGranularity());
    // Get both granules
    Granule i = alpha.getGranule();
    Granule j = gamma.getGranule();
    return i.compareTo(j) < 0;
    }

  /** 
  * Does an Instant precede the start of a Period, 
  * that is, is alpha < beta.start?
  * @param alpha - Instant to compare
  * @param beta - Period to compare
  * @return alpha < beta.start
 * @throws TemporalDataTypeFormationException 
 * @throws CalendricSystemServiceException 
  **/
  public boolean precedes(Instant alpha, Period beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
    // Convert beta to alpha's granularity
    Instant gamma = beta.getStarting();
    gamma = gamma.cast(alpha.getGranularity());

    // Get both granules
    Granule i = alpha.getGranule();
    Granule j = gamma.getGranule();
    return i.compareTo(j) < 0;
    }

  /** 
  * Does the end of a Period precede the start of an Instant, 
  * that is, is alpha.end < beta?
  * @param alpha - Period to compare
  * @param beta - Instant to compare
  * @return alpha.end < beta
 * @throws TemporalDataTypeFormationException 
 * @throws CalendricSystemServiceException 
  **/
  public boolean precedes(Period alpha, Instant beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
    // Convert beta to alpha's granularity
    Instant alphaEnd = alpha.getEnding();
    Instant gamma = beta.cast(alphaEnd.getGranularity());

    // Get both granules
    Granule i = alphaEnd.getGranule();
    Granule j = gamma.getGranule();
    return i.compareTo(j) < 0;
    }

  /** 
  * Does the end of a Period precede the start of another,
  * that is, is alpha.end < beta.start?
  * @param alpha - Period to compare
  * @param beta - Period to compare
  * @return alpha.end < beta.start
 * @throws TemporalDataTypeFormationException 
 * @throws CalendricSystemServiceException 
  **/
  public boolean precedes(Period alpha, Period beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
    // Convert beta to alpha's granularity
    Instant alphaEnd = alpha.getEnding();
    Instant betaStart = beta.getStarting();
    Instant gamma = betaStart.cast(alphaEnd.getGranularity());

    // Get both granules
    Granule i = alphaEnd.getGranule();
    Granule j = gamma.getGranule();
    return i.compareTo(j) < 0;
    }

  /** 
  * Does this Period begin at the same time as another, but end before it,
  * that is, is alpha.start == beta.start but alpha.end < beta.end?  
  * @param alpha - Period to compare
  * @param beta - Period to compare
  * @return alpha.start == beta.start AND alpha.end < beta.end
 * @throws TemporalDataTypeFormationException 
 * @throws CalendricSystemServiceException 
  **/
  public boolean starts(Period alpha, Period beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
    // Convert beta to alpha's granularity
    Instant alphaEnd = alpha.getEnding();
    Instant alphaStart = alpha.getStarting();
    Instant betaStart = beta.getStarting();
    Instant betaEnd = beta.getEnding();
    Instant gammaStart = betaStart.cast(alphaStart.getGranularity());
    Instant gammaEnd = betaEnd.cast(alphaEnd.getGranularity());

    // Get both granules
    Granule i = alphaStart.getGranule();
    Granule j = gammaStart.getGranule();
    Granule k = alphaEnd.getGranule();
    Granule m = gammaEnd.getGranule();
    return i.compareTo(j) == 0 && k.compareTo(m) < 0;
  }

  /** 
  * Does this Period finish at the same time as another, but start after it,
  * that is, is alpha.end == beta.end but alpha.start < beta.start?  
  * @param alpha - Period to compare
  * @param beta - Period to compare
  * @return alpha.end == beta.end AND alpha.start < beta.start
 * @throws TemporalDataTypeFormationException 
 * @throws CalendricSystemServiceException 
  **/
  public boolean finishes(Period alpha, Period beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
    // Convert beta to alpha's granularity
    Instant alphaEnd = alpha.getEnding();
    Instant alphaStart = alpha.getStarting();
    Instant betaStart = beta.getStarting();
    Instant betaEnd = beta.getEnding();
    Instant gammaStart = betaStart.cast(alphaStart.getGranularity());
    Instant gammaEnd = betaEnd.cast(alphaEnd.getGranularity());

    // Get both granules
    Granule i = alphaStart.getGranule();
    Granule j = gammaStart.getGranule();
    Granule k = alphaEnd.getGranule();
    Granule m = gammaEnd.getGranule();
    return i.compareTo(j) < 0 && k.compareTo(m) == 0;
  }

  /** 
  * Does the end of a Period meet the start of another Period,
  * that is, is alpha.end == beta.start?  This is the inverse of metBy().
  * @param alpha - Period to compare
  * @param beta - Period to compare
  * @return alpha.end == beta.start
 * @throws TemporalDataTypeFormationException 
 * @throws CalendricSystemServiceException 
  **/
  public boolean meets(Period alpha, Period beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
    // Convert beta to alpha's granularity
    Instant alphaEnd = alpha.getEnding();
    Instant betaStart = beta.getStarting();
    Instant gamma = betaStart.cast(alphaEnd.getGranularity());

    // Get both granules
    Granule i = alphaEnd.getGranule();
    Granule j = gamma.getGranule();
    return i.compareTo(j) == 0;
  }

  /** 
  * Is the start of a Period metBy the start of another Period,
  * that is, is beta.end == alpha.start?  This is the inverse of meets().
  * @param alpha - Period to compare
  * @param beta - Period to compare
  * @return alpha.start == beta.end
 * @throws TemporalDataTypeFormationException 
 * @throws CalendricSystemServiceException 
  **/
  public boolean metBy(Period alpha, Period beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
    // Convert beta to alpha's granularity
    Instant alphaStart = alpha.getStarting();
    Instant betaEnd = beta.getEnding();
    Instant gamma = betaEnd.cast(alphaStart.getGranularity());

    // Get both granules
    Granule i = alphaStart.getGranule();
    Granule j = gamma.getGranule();
    return i.compareTo(j) == 0;
  }

  /** 
  * Does one Period overlap another?  Two Periods overlap if they intersect
  * in time.  This operation is symmetric.
  * @param alpha - Period to compare
  * @param beta - Period to compare
  * @return alpha overlaps beta
 * @throws TemporalDataTypeFormationException 
 * @throws CalendricSystemServiceException 
  **/
  public boolean overlaps(Period alpha, Period beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
    // Convert beta to alpha's granularity
    Instant alphaEnd = alpha.getEnding();
    Instant alphaStart = alpha.getStarting();
    Instant betaStart = beta.getStarting();
    Instant betaEnd = beta.getEnding();
    Instant gammaStart = betaStart.cast(alphaStart.getGranularity());
    Instant gammaEnd = betaEnd.cast(alphaEnd.getGranularity());

    // Get both granules
    Granule i = alphaStart.getGranule();
    Granule j = gammaStart.getGranule();
    Granule k = alphaEnd.getGranule();
    Granule m = gammaEnd.getGranule();
    return !(k.compareTo(j) < 0 || m.compareTo(i) < 0);
  }

  /** 
  * Does this Period contain another?  A periods contains another if
  * it starts before it and ends after it.  The containment need not
  * be strict.
  * @param alpha - Period to compare
  * @param beta - Period to compare
  * @return alpha.start <= beta.start AND alpha.end >= beta.end
 * @throws TemporalDataTypeFormationException 
 * @throws CalendricSystemServiceException 
  **/
  public boolean contains(Period alpha, Period beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
    // Convert beta to alpha's granularity
    Instant alphaEnd = alpha.getEnding();
    Instant alphaStart = alpha.getStarting();
    Instant betaStart = beta.getStarting();
    Instant betaEnd = beta.getEnding();
    Instant gammaStart = betaStart.cast(alphaStart.getGranularity());
    Instant gammaEnd = betaEnd.cast(alphaEnd.getGranularity());

    // Get both granules
    Granule i = alphaStart.getGranule();
    Granule j = gammaStart.getGranule();
    Granule k = alphaEnd.getGranule();
    Granule m = gammaEnd.getGranule();
    return i.compareTo(j) <= 0 && k.compareTo(m) >= 0;
  }

  /** 
  * Is this Period during another?  A periods is during another if
  * it starts after it and ends before it.  The containment need not
  * be strict. 
  * @param alpha - Period to compare
  * @param beta - Period to compare
  * @return alpha.start >= beta.start AND alpha.end <= beta.end
 * @throws TemporalDataTypeFormationException 
 * @throws CalendricSystemServiceException 
  **/
  public boolean during(Period alpha, Period beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
    // Convert beta to alpha's granularity
    Instant alphaEnd = alpha.getEnding();
    Instant alphaStart = alpha.getStarting();
    Instant betaStart = beta.getStarting();
    Instant betaEnd = beta.getEnding();
    Instant gammaStart = betaStart.cast(alphaStart.getGranularity());
    Instant gammaEnd = betaEnd.cast(alphaEnd.getGranularity());

    // Get both granules
    Granule i = alphaStart.getGranule();
    Granule j = gammaStart.getGranule();
    Granule k = alphaEnd.getGranule();
    Granule m = gammaEnd.getGranule();
    return i.compareTo(j) >= 0 && k.compareTo(m) <= 0;
  }

  /** 
  * Does this Period represent the same times as another?  
  * A period is equalTo another if it both starts and finishes it.
  * This operation is symmetric.
  * @param alpha - Period to compare
  * @param beta - Period to compare
  * @return alpha.start == beta.start AND alpha.end == beta.end
 * @throws TemporalDataTypeFormationException 
 * @throws CalendricSystemServiceException 
  **/
  public boolean equalTo(Period alpha, Period beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
    // Convert beta to alpha's granularity
    Instant alphaEnd = alpha.getEnding();
    Instant alphaStart = alpha.getStarting();
    Instant betaStart = beta.getStarting();
    Instant betaEnd = beta.getEnding();
    Instant gammaStart = betaStart.cast(alphaStart.getGranularity());
    Instant gammaEnd = betaEnd.cast(alphaEnd.getGranularity());

    // Get both granules
    Granule i = alphaStart.getGranule();
    Granule j = gammaStart.getGranule();
    Granule k = alphaEnd.getGranule();
    Granule m = gammaEnd.getGranule();
    return i.compareTo(j) == 0 && k.compareTo(m) == 0;
  }

  /** 
  * Does this Instant represent the same times as another?  
  * An instant is equalTo another if it is at the same time.
  * This operation is symmetric.
  * @param alpha - Instant to compare
  * @param beta - Instant to compare
  * @return alpha == beta
 * @throws TemporalDataTypeFormationException 
 * @throws CalendricSystemServiceException 
  **/
  public boolean equalTo(Instant alpha, Instant beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
    // Convert beta to alpha's granularity
    Instant gamma = beta.cast(alpha.getGranularity());

    // Get both granules
    Granule i = alpha.getGranule();
    Granule j = gamma.getGranule();
    return i.compareTo(j) == 0;
  }

  /** 
  * Displace (add) a Period on the time-line by an Interval 
  * @param alpha - displacee
  * @param beta - displacer
  * @return - [alpha.start + beta, alpha.end + beta]
 * @throws TemporalDataTypeFormationException 
 * @throws CalendricSystemServiceException 
  **/
  public Period add(Period alpha, Interval beta) throws TemporalDataTypeFormationException, CalendricSystemServiceException {
    // Convert beta to alpha's granularity
    Instant alphaEnd = alpha.getEnding();
    Instant alphaStart = alpha.getStarting();
    Interval gamma = beta.cast(alpha.getGranularity());

    // Get both granules
    Granule i = alphaStart.getGranule();
    Granule j = gamma.smallest();
    Granule k = alphaEnd.getGranule();
    Granule m = gamma.largest();
    Granule low = i.add(j);
    Granule high = k.add(m);
    return new Period(alpha.getGranularity(), 
                      new Instant(low), new Instant(high));
  }

  /** 
  * Displace (add) an Instant on the time-line by an Interval 
  * @param alpha - displacee
  * @param beta - displacer
  * @return - alpha + beta
 * @throws CalendricSystemServiceException 
 * @throws TemporalDataTypeFormationException 
  **/
  public Instant add(Instant alpha, Interval beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
    // Convert beta to alpha's granularity
    Interval gamma = beta.cast(alpha.getGranularity());

    // Get both granules
    Granule i = alpha.getGranule();
    Granule j = gamma.smallest();
    Granule g = i.add(j);
    return new Instant(alpha.getGranularity(), g);
  }

  /** 
  * Add one Interval to another
  * @param alpha - displacee
  * @param beta - displacer
  * @return - alpha + beta
 * @throws CalendricSystemServiceException 
  **/
  public Interval add(Interval alpha, Interval beta) throws CalendricSystemServiceException {
    // Convert beta to alpha's granularity
    Interval gamma = beta.cast(alpha.getGranularity());

    // Get both granules
    Granule i = alpha.smallest();
    Granule j = gamma.smallest();
    Granule g = i.add(j);
    return new Interval(alpha.getGranularity(), g);
  }

  /** 
  * Displace (subtract) a Period on the time-line by an Interval 
  * @param alpha - displacee
  * @param beta - displacer
  * @return - [alpha.start - beta, alpha.end - beta]
 * @throws CalendricSystemServiceException 
 * @throws TemporalDataTypeFormationException 
  **/
  public Period subtract(Period alpha, Interval beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
    // Convert beta to alpha's granularity
    Instant alphaEnd = alpha.getEnding();
    Instant alphaStart = alpha.getStarting();
    Interval gamma = beta.cast(alpha.getGranularity());

    // Get both granules
    Granule i = alphaStart.getGranule();
    Granule j = gamma.smallest();
    Granule k = alphaEnd.getGranule();
    Granule m = gamma.largest();
    Granule low = i.subtract(j);
    Granule high = k.subtract(m);
    return new Period(alpha.getGranularity(), 
                      new Instant(low), new Instant(high));
  }

  /** 
  * Displace (subtract) an Instant on the time-line by an Interval 
  * @param alpha - displacee
  * @param beta - displacer
  * @return - alpha - beta
 * @throws CalendricSystemServiceException 
 * @throws TemporalDataTypeFormationException 
  **/
  public Instant subtract(Instant alpha, Interval beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
    // Convert beta to alpha's granularity
    Interval gamma = beta.cast(alpha.getGranularity());

    // Get both granules
    Granule i = alpha.getGranule();
    Granule j = gamma.smallest();
    Granule g = i.subtract(j);
    return new Instant(alpha.getGranularity(), g);
  }

  /** 
  * Subtract one Interval from another
  * @param alpha - displacee
  * @param beta - displacer
  * @return - alpha - beta
 * @throws CalendricSystemServiceException 
  **/
  public Interval subtract(Interval alpha, Interval beta) throws CalendricSystemServiceException {
    // Convert beta to alpha's granularity
    Interval gamma = beta.cast(alpha.getGranularity());

    // Get both granules
    Granule i = alpha.smallest();
    Granule j = gamma.smallest();
    Granule g = i.subtract(j);
    return new Interval(alpha.getGranularity(), g);
  }

  /** 
  * Negate an Interval. 
  * @param alpha - Interval to negate
  * @return - -alpha
  **/
  public Interval negate(Interval alpha) {
    // Get both granules
    Granule i = alpha.smallest();
    return new Interval(alpha.getGranularity(), i.negate());
  }

  /** 
  * Multiply an Interval by an integer constant.  
  * @param alpha - Interval to multiply
  * @param n - by how many 
  * @return result - (alpha * n)
  **/
  public Interval multiply(Interval alpha, int n) {
    // Get both granules
    Granule i = alpha.smallest();
    return new Interval(alpha.getGranularity(), i.multiply(n));
   }

  /** 
  * Divide an Interval by an integer constant.  
  * @param alpha - Interval to divide
  * @param n - by how many 
  * @return result - (alpha div n)
  **/
  public Interval divide(Interval alpha, int n) {
    // Get both granules
    Granule i = alpha.smallest();
    return new Interval(alpha.getGranularity(), i.divide(n));
  }

  /** 
  * Divide an Interval by an Interval (integer division). 
  * @param alpha - divisee
  * @param beta - divisor
  * @param result - alpha div beta
  * @status NOT IMPLEMENTED - problematic for MAX/MIN
  **/
  public int divide(Interval alpha, Interval beta) {
    return 0;
  }

  /** 
  * Temporal constructor - Choose the first (earliest) from among a pair of instants
  * @param alpha - Instant 
  * @param beta - Instant 
  * @return first Instant on the time-line from among alpha and beta
 * @throws TemporalDataTypeFormationException 
 * @throws CalendricSystemServiceException 
  **/
  public Instant first(Instant alpha, Instant beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
    // Convert beta to alpha's granularity
    Instant gamma = beta.cast(alpha.getGranularity());

    Granule i = alpha.getGranule();
    Granule j = gamma.getGranule();
    if (i.compareTo(j) < 0) { return alpha; }
    return gamma;
  }

  /** 
  * Temporal constructor - Choose the last (latest) from among a pair of instants
  * @param alpha - Instant 
  * @param beta - Instant 
  * @return last Instant on the time-line from among alpha and beta
 * @throws TemporalDataTypeFormationException 
 * @throws CalendricSystemServiceException 
  **/
  public Instant last(Instant alpha, Instant beta) throws CalendricSystemServiceException, TemporalDataTypeFormationException {
    // Convert beta to alpha's granularity
    Instant gamma = beta.cast(alpha.getGranularity());

    Granule i = alpha.getGranule();
    Granule j = gamma.getGranule();
    if (i.compareTo(j) < 0) { return gamma; }
    return alpha;
  }

}